/*************************************************************************/
/**                                                                     **/
/**  cfa631 - Crystalfontz CFA-631 USB LCD module utility               **/
/**                                                                     **/
/**  Greg Bailey <gebailey@inter-tel.com>                               **/
/**  November 13, 2005                                                  **/
/**                                                                     **/
/**  cfa631 Copyright (C) 2005 Inter-Tel, Inc.                          **/
/**  Copyright FontzControl 2005 Keven Tipping                          **/
/**  Copyright Cfont631Lib 2004 Guillermo Alcaraz                       **/
/**                                                                     **/
/**  cfa631/FontzControl/cfont631lib is free software; you can          **/
/**  redistribute it and/or modify it under the terms of the GNU        **/
/**  General Public License as published by the Free Software           **/
/**  Foundation; either version 2 of the License, or (at your option)   **/
/**  any later version.                                                 **/
/**                                                                     **/
/**  cfa631/FontzControl/cfont631lib is distributed in the hope that    **/
/**  it will be useful, but WITHOUT ANY WARRANTY; without even the      **/
/**  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR    **/
/**  PURPOSE.  See the GNU General Public License for more details.     **/
/**                                                                     **/
/**  You should have received a copy of the GNU General Public License  **/
/**  along with this program; if not, write to the Free Software        **/
/**  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA          **/
/**  02111-1307, USA                                                    **/
/**                                                                     **/
/**  cfa631 is based on "FontzControl", which is used to control a      **/
/**  Crystalfontz CFA-631 USB LCD module.  "FontzControl" was           **/
/**  downloaded from the CrystalFontz forums web site.                  **/
/**                                                                     **/
/**  CONFIGURATION: Please change the location SERIAL_PORT to wherever  **/
/**  your display is. (usually ucom0 for FreeBSD, ttyUSB0 for Linux,    **/
/**  etc). "Dmesg | Grep fontz" should tell you where the display is    **/
/**  hooked into through Dev if you have the FTDI modules installed in  **/
/**  your kernel. Note that FreeBSD users will have to put              **/
/**  uftdi_load="YES" in their Boot Configuration to load the FTDI      **/
/**  driver up before the UGEN driver grabs the display.  In Linux,     **/
/**  you'll need to "usbserial" and "ftdi_sio" modules loaded.          **/
/**                                                                     **/
/*************************************************************************/

/*************************************************************************/
/**  INCLUDES                                                           **/
/*************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <pthread.h>
#include <termios.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/time.h>
#include <sys/types.h>


/*************************************************************************/
/**  TYPEDEFS                                                           **/
/*************************************************************************/

typedef int                int32;
typedef unsigned short int uint16;
typedef unsigned char      uchar8;
typedef unsigned char      uint8;
typedef unsigned int       uint32;

typedef union {
        uint16  word;
        uint8 byte[2];
} WORD;

/* All packets have the following structure:
 *
 * <type><data_length><data><CRC>
 *
 * type is a bitfield with the following values
 *
 * 0x00 - command from host to CFA-631
 * 0x40 - ack/report from CFA-631 to host
 * 0x80 - unsolicited message from CFA-631 to host
 * 0xc0 - command error
 *
 * lower 6 bits of type is the command.
 *
 * data_length value 0-22.
 *
 * CRC is a 16 bit low-endian checksum.
 */
typedef struct {
        uint8 type;
        uint8 length;
        uint8 info[24];
} LCD_PACKET_DATA;

typedef struct {
        int type;               /* type of event: 0 keypress, 1 keyrelease */
        int keycode;            /* keycode of the key pressed or released */
} CFONT631_KEY_EVENT;

typedef union {
        int type;                       /* type of event */
        CFONT631_KEY_EVENT key;         /* keyevents */
} CFONT631_EVENT;

typedef union {
        LCD_PACKET_DATA packet;
        uint8 buffer[256];
} LCD_SEND_DATA;

typedef union {
        LCD_PACKET_DATA packet;
        uint8 buffer[256];
} LCD_RECEIVE_DATA;

#define BUFFER_MAX_SIZE 256

typedef struct {
        LCD_SEND_DATA *buffer[BUFFER_MAX_SIZE];
        int len;
        int tail;
        int head;
        void (*event_handler) (CFONT631_EVENT *e, void *info);
        void *info;
        int reset_after_end;
} LCD_SEND_BUFFER;

/*************************************************************************/
/**  CONSTANTS                                                          **/
/*************************************************************************/

/* Event Types */
#define CFONT631_EVENT_KEYPRESS   0
#define CFONT631_EVENT_KEYRELEASE 1

/* Keycodes */
#define CFONT631_KEY_UP_LEFT    1
#define CFONT631_KEY_UP_RIGHT   2
#define CFONT631_KEY_DOWN_LEFT  3
#define CFONT631_KEY_DOWN_RIGHT 4

/* Cursor Styles */
#define CFONT631_CURSOR_STYLE_NO_CURSOR         0
#define CFONT631_CURSOR_STYLE_BLINK             1
#define CFONT631_CURSOR_STYLE_UNDERSCORE        2
#define CFONT631_CURSOR_STYLE_BLINK_UNDER       3
#define CFONT631_CURSOR_STYLE_INVERTING_BLINK   4

/* Key Legends */
#define CFONT631_KEY_LEGEND_BLANK       0
#define CFONT631_KEY_LEGEND_CANCEL      1
#define CFONT631_KEY_LEGEND_CHECK       2
#define CFONT631_KEY_LEGEND_UP          3
#define CFONT631_KEY_LEGEND_DOWN        4
#define CFONT631_KEY_LEGEND_RIGHT       5
#define CFONT631_KEY_LEGEND_LEFT        6
#define CFONT631_KEY_LEGEND_PLUS        7
#define CFONT631_KEY_LEGEND_MINUS       8
#define CFONT631_KEY_LEGEND_NONE        9

/*************************************************************************/
/**  FUNCTION PROTOTYPES                                                **/
/*************************************************************************/

extern "C" {
int cfont631_open (const char *device);
void cfont631_reset ();
void cfont631_close ();
void cfont631_reset_after_end (const int on);
void cfont631_reboot ();
void cfont631_clear ();
void cfont631_set_cursor_position (const int x, const int y);
void cfont631_set_cursor_style (const int style);
void cfont631_set_contrast (const int level_percent);
void cfont631_set_backlight (const int level_percent);
void cfont631_write_xy (const int x, const int y, const char *s, const int len);
void cfont631_write_char_xy (const int x, const int y, const uint8 c);
void cfont631_define_char (const int index, const uint8 matrix[8]);
void cfont631_enable_key_legend (const int kul, const int kur, const int kll, const int klr);
void cfont631_disable_key_legend ();
void cfont631_store_boot_state ();
void cfont631_register_event_handler (void (*event_handler) (CFONT631_EVENT *e, void *info), void *info);
}

/* eof */
