/*************************************************************************/
/**                                                                     **/
/**  cfa631 - Crystalfontz CFA-631 USB LCD module utility               **/
/**                                                                     **/
/**  Greg Bailey <gebailey@inter-tel.com>                               **/
/**  November 13, 2005                                                  **/
/**                                                                     **/
/**  cfa631 Copyright (C) 2005 Inter-Tel, Inc.                          **/
/**  Copyright FontzControl 2005 Keven Tipping                          **/
/**  Copyright Cfont631Lib 2004 Guillermo Alcaraz                       **/
/**                                                                     **/
/**  cfa631/FontzControl/cfont631lib is free software; you can          **/
/**  redistribute it and/or modify it under the terms of the GNU        **/
/**  General Public License as published by the Free Software           **/
/**  Foundation; either version 2 of the License, or (at your option)   **/
/**  any later version.                                                 **/
/**                                                                     **/
/**  cfa631/FontzControl/cfont631lib is distributed in the hope that    **/
/**  it will be useful, but WITHOUT ANY WARRANTY; without even the      **/
/**  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR    **/
/**  PURPOSE.  See the GNU General Public License for more details.     **/
/**                                                                     **/
/**  You should have received a copy of the GNU General Public License  **/
/**  along with this program; if not, write to the Free Software        **/
/**  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA          **/
/**  02111-1307, USA                                                    **/
/**                                                                     **/
/**  cfa631 is based on "FontzControl", which is used to control a      **/
/**  Crystalfontz CFA-631 USB LCD module.  "FontzControl" was           **/
/**  downloaded from the CrystalFontz forums web site.                  **/
/**                                                                     **/
/**  CONFIGURATION: Please change the location SERIAL_PORT to wherever  **/
/**  your display is. (usually ucom0 for FreeBSD, ttyUSB0 for Linux,    **/
/**  etc). "Dmesg | Grep fontz" should tell you where the display is    **/
/**  hooked into through Dev if you have the FTDI modules installed in  **/
/**  your kernel. Note that FreeBSD users will have to put              **/
/**  uftdi_load="YES" in their Boot Configuration to load the FTDI      **/
/**  driver up before the UGEN driver grabs the display.  In Linux,     **/
/**  you'll need to "usbserial" and "ftdi_sio" modules loaded.          **/
/**                                                                     **/
/*************************************************************************/

/*************************************************************************/
/**  INCLUDES                                                           **/
/*************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <pthread.h>
#include <termios.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/time.h>
#include <sys/types.h>

#include "cfa631.h"

/*************************************************************************/
/**  CONFIGURATION                                                      **/
/*************************************************************************/

#define SERIAL_PORT "/dev/usb/tts/0"



static void define_bignums (void);
static void bignum_write_digit_x( const int x, int num );
static void bignum_write_text_x( const int x, const char *s );


/*************************************************************************/
/**  USAGE                                                              **/
/*************************************************************************/

void usage( char *argv[] )
{
    printf( "\n" );
    printf( "cfa631 - Crystalfontz CFA-631 USB LCD module utility\n" );
    printf( "\n" );
    printf( "Usage: %s COMMAND OPTION1 OPTION2...\n", argv[0] );
    printf( "\n" );
    printf( "       cursor OPTION1 : Set the cursor. 0=No Cursor, 1=Blink, 2=Underscore,\n" );
    printf( "           3=Blink Underscore, 4=Inverting Blink\n" );
    printf( "       getkey OPTION1 OPTION2 : Retrieves a key event from the CFA-631. When\n" );
    printf( "           OPTION1 is 0, the program will wait OPTION2 seconds until exiting. If\n" );
    printf( "           no key is pressed, program will print 0. If OPTION1 is 1, the program\n" );
    printf( "           will wait indefinitely until a key is pressed, then print the code of\n" );
    printf( "           the key. Key Codes are as follows:\n" );
    printf( "           1 = Upper Left PRESS     10 = Upper Left RELEASE\n" );
    printf( "           2 = Upper Right PRESS    20 = Upper Right RELEASE\n" );
    printf( "           3 = Lower Left PRESS     30 = Lower Left RELEASE\n" );
    printf( "           4 = Lower Right PRESS    40 = Lower Right RELEASE\n" );
    printf( "       keylegend OPTION1 OPTION2 OPTION3 OPTION4 : Set the desired keylegend for\n" );
    printf( "           OPTION1, where the option is 0=Blank, 1=Cancel, 2=Check, 3=Up,\n" );
    printf( "           4=Down, 5=Right, 6=Left, 7=Plus, 8=Minus, 9=Empty, Use in format UL\n" );
    printf( "           UR BL BR, so 1 2 3 4 would set UL=Cancel, UR=Check, BL=Up, BR=Down.\n" );
    printf( "       keydisable : Disable the keylegend\n" );
    printf( "       backlight OPTION1 : Set the backlight to 0-99 for OPTION1.\n" );
    printf( "       contrast OPTION1 : Set the contrast to 0-100 for OPTION1.\n" );
    printf( "       reset : Reset the 631 Unit.\n" );
    printf( "       reboot : Reboot the 631 Unit.\n" );
    printf( "       cursorpos OPTION1 OPTION2 : Set the cursor to X (OPTION1), Y(OPTION2)\n" );
    printf( "       storeboot : Store the CURRENT display contents to boot ROM.\n" );
    printf( "       clear : Clear the contents of the display\n" );
    printf( "       writetext OPTION1 OPTION2 'OPTION3' : Write text to X (OPTION1),\n" );
    printf( "           Y (OPTION2), and Text to be displayed as 'OPTION3'.\n" );
    printf( "       writechar OPTION1 OPTION2 OPTION3 : Write a charactor as X (OPTION1),\n" );
    printf( "           Y (OPTION2), and the Charactor number as OPTION3.\n" );
    printf( "       menu OPTION1 OPTION2 ... : Write a menu.  Returns the index of the\n" );
    printf( "           option selected; 'cancel' if user pressed 'cancel'.\n" );
    printf( "       ip OPTION1 : Edit an IP address. OPTION1 must be of form\n" );
    printf( "           'DDD.DDD.DDD.DDD'. Returns modified IP address or 'cancel'.\n" );
    printf( "\n" );
}

static int fin;

void handler (CFONT631_EVENT *e, void *info) {
        if (e->type == CFONT631_EVENT_KEYPRESS) {
                if (e->key.keycode == CFONT631_KEY_UP_LEFT) {
                        fin = 1;
                } else if(e->key.keycode == CFONT631_KEY_UP_RIGHT) {
                        fin = 2;
                } else if(e->key.keycode == CFONT631_KEY_DOWN_LEFT) {
                        fin = 3;
                } else if(e->key.keycode == CFONT631_KEY_DOWN_RIGHT) {
                        fin = 4;
                }
        }
        if (e->type == CFONT631_EVENT_KEYRELEASE) {
                if (e->key.keycode == CFONT631_KEY_UP_LEFT) {
                        fin = 10;
                } else if(e->key.keycode == CFONT631_KEY_UP_RIGHT) {
                        fin = 20;
                } else if(e->key.keycode == CFONT631_KEY_DOWN_LEFT) {
                        fin = 30;
                } else if(e->key.keycode == CFONT631_KEY_DOWN_RIGHT) {
                        fin = 40;
                }
        }
}

int comparestr(char string1[255], char string2[255])
{
        int found=0,counter=0;
        while(counter<strlen(string1) && found < strlen(string1))
        {
                if(string1[found]==string2[counter])
                {
                        found++;
                } else
                {
                        found=0;
                }
                counter++;
        }
        if(found==strlen(string1))
        {
                return 1;
        } else
        {
                return -1;
        }
}

/*************************************************************************/
/**  runcommands()                                                      **/
/*************************************************************************/

void runcommands( int argc, char *args[] )
{
    int opt1 = 0, opt2 = 0, opt3 = 0, opt4 = 0;

    if ( comparestr( args[1], "cursor" ) > 0 )
    {
        if ( argc == 2 || atoi( args[2] ) < 0 || atoi( args[2] ) > 4 )
        {
            printf( "Error, invalid or no option!\n" );
            return;
        }
        opt1 = atoi( args[2] );
        cfont631_set_cursor_style( opt1 );
        return;
    }
    else if ( comparestr( args[1], "getkey" ) > 0 )
    {
        if ( argc != 4 || atoi( args[2] ) < 0 || atoi( args[2] ) > 1 )
        {
            printf( "Error, invalid or no option!\n" );
            return;
        }
        fin = 0;
        cfont631_register_event_handler( handler, NULL );
        opt1 = atoi( args[2] );
        opt2 = atoi( args[3] );
        if ( opt1 == 1 )
        {
            while ( fin == 0 )
            {
                usleep( 10000 );
            }
            printf( "%d", fin );
            return;
        }
        else
        {
            clock_t endwait;
            endwait = time( (time_t *)0 ) + opt2;
            while ( time( (time_t *)0 ) < endwait && fin == 0 )
            {
                usleep( 10000 );
            }
            printf( "%d", fin );
            return;
        }
        return;
    }
    else if ( comparestr( args[1], "keylegend" ) > 0 )
    {
        if ( argc != 6 || atoi( args[2] ) < 0 || atoi( args[2] ) > 9
                       || atoi( args[3] ) < 0 || atoi( args[3] ) > 9
                       || atoi( args[4] ) < 0 || atoi( args[4] ) > 9
                       || atoi( args[5] ) < 0 || atoi( args[5] ) > 9 )
        {
            printf( "Error, invalid or no option!\n" );
            return;
        }
        opt1 = atoi( args[2] );
        opt2 = atoi( args[3] );
        opt3 = atoi( args[4] );
        opt4 = atoi( args[5] );
        cfont631_enable_key_legend( opt1, opt2, opt3, opt4 );
        return;
    }
    else if ( comparestr( args[1], "keydisable" ) > 0 )
    {
        cfont631_disable_key_legend();
        return;
    }
    else if ( comparestr( args[1], "backlight" ) > 0 )
    {
        if ( argc == 2 || atoi( args[2] ) < 0 || atoi( args[2] ) > 99 )
        {
            printf( "Error, invalid or no option!\n" );
            return;
        }
        opt1 = atoi( args[2] );
        cfont631_set_backlight( opt1 );
        return;
    }
    else if ( comparestr( args[1], "contrast" ) > 0 )
    {
        if ( argc == 2 || atoi( args[2] ) < 0 || atoi( args[2] ) > 99 )
        {
            printf( "Error, invalid or no option!\n" );
            return;
        }
        opt1 = atoi( args[2] );
        cfont631_set_contrast( opt1 );
        return;
    }
    else if ( comparestr( args[1], "reset" ) > 0 )
    {
        cfont631_reset();
        return;
    }
    else if ( comparestr( args[1], "reboot" ) > 0 )
    {
        cfont631_reboot();
        return;
    }
    else if ( comparestr( args[1], "cursorpos" ) > 0 )
    {
        if ( argc != 4 || atoi( args[2] ) < 1 || atoi( args[2] ) > 20
                       || atoi( args[3] ) < 1 || atoi( args[3] ) > 2 )
        {
            printf( "Error, invalid or no option!\n" );
            return;
        }
        opt1 = atoi( args[2] );
        opt2 = atoi( args[3] );
        cfont631_set_cursor_position( opt1, opt2 );
        return;
    }
    else if ( comparestr( args[1], "storeboot" ) > 0 )
    {
        cfont631_store_boot_state();
        return;
    }
    else if ( comparestr( args[1], "clear" ) > 0 )
    {
        cfont631_clear();
        return;
    }
    else if ( comparestr( args[1], "writetext" ) > 0 )
    {
        if ( argc != 5 || atoi( args[2] ) < 1 || atoi( args[2] ) > 20
                       || atoi( args[3] ) < 1 || atoi( args[3] ) > 2 )
        {
            printf( "Error, invalid or no option!\n" );
            return;
        }
        opt1 = atoi( args[2] );
        opt2 = atoi( args[3] );
        cfont631_write_xy( opt1, opt2, args[4], strlen(args[4]) );
        return;
    }
    else if ( comparestr( args[1], "writechar" ) > 0 )
    {
        if ( argc != 5 || atoi( args[2] ) < 1 || atoi( args[2] ) > 20
                       || atoi( args[3] ) < 1 || atoi( args[3] ) > 2 )
        {
            printf( "Error, invalid or no option!\n" );
            return;
        }
        opt1 = atoi( args[2] );
        opt2 = atoi( args[3] );
        opt3 = atoi( args[4] );
        cfont631_write_char_xy( opt1, opt2, opt3 );
        return;
    }
    else if ( comparestr( args[1], "definebignum" ) > 0 )
    {
    	define_bignums();
	return;
    }
    else if ( comparestr( args[1], "writebignum" ) > 0 )
    {
        if ( argc != 4 || atoi( args[2] ) < 1 || atoi( args[2] ) > 20 )
        {
            printf( "Error, invalid or no option!\n" );
            return;
        }
    	opt1 = atoi( args[2] );
    	bignum_write_text_x( opt1, args[3] );
	return;
    }
    else if ( comparestr( args[1], "menu" ) > 0 )
    {
        int current_index = 1;
        int current_line = 1;
        if ( argc < 3 )
        {
            printf( "Error, invalid or no option!\n" );
            return;
        }
        cfont631_enable_key_legend ( CFONT631_KEY_LEGEND_UP, CFONT631_KEY_LEGEND_CHECK,
                                     CFONT631_KEY_LEGEND_DOWN, CFONT631_KEY_LEGEND_CANCEL );
        cfont631_clear();
        cfont631_write_char_xy( 1, current_line, 16 );
        cfont631_write_xy( 2, current_line, args[current_index+1], strlen(args[current_index+1]) );
        if ( ( current_index + 2 ) < argc )
        {
            cfont631_write_xy( 2, 2, args[current_index+2], strlen(args[current_index+2]) );
        }
        cfont631_register_event_handler( handler, NULL );
        while ( 1 )
        {
            fin = 0;
            while ( fin == 0 )
            {
                usleep( 10000 );
            }
            if ( fin == CFONT631_KEY_UP_LEFT )
            {
                current_index--;
                if ( current_index < 1 ) current_index = 1;
                current_line--;
                if ( current_line < 1 ) current_line = 1;
            }
            else if ( fin == CFONT631_KEY_UP_RIGHT )
            {
                printf( "%d\n", current_index );
                break;
            }
            else if ( fin == CFONT631_KEY_DOWN_LEFT )
            {
                current_index++;
                if ( current_index > ( argc - 2 ) ) current_index = ( argc - 2 );
                current_line++;
                if ( current_line > 2 ) current_line = 2;
            }
            else if ( fin == CFONT631_KEY_DOWN_RIGHT )
            {
                printf( "cancel\n" );
                break;
            }
            cfont631_clear();
            cfont631_write_char_xy( 1, current_line, 16 );
            cfont631_write_xy( 2, current_line, args[current_index+1], strlen(args[current_index+1]) );
            if ( ( current_line == 1 ) && ( ( current_index + 2 ) < argc ) )
            {
                cfont631_write_xy( 2, 2, args[current_index+2], strlen(args[current_index+2]) );
            }
            else if ( ( current_line == 2 ) && ( current_index > 1 ) )
            {
                cfont631_write_xy( 2, 1, args[current_index], strlen(args[current_index]) );
            }
        }
        cfont631_disable_key_legend();
        return;
    }
    else if ( comparestr( args[1], "ip" ) > 0 )
    {
        char ip_addr[16];
        int current_digit = 1;
        int current_octet;
        char ascii_octet[4];
        if ( argc != 3 )
        {
            printf( "Error, invalid or no option!\n" );
            return;
        }
        strncpy( ip_addr, args[2], 15 );
        ip_addr[15] = '\0';
        if ( ! ( ( isdigit( ip_addr[0] ) ) && ( isdigit( ip_addr[1] ) ) && ( isdigit( ip_addr[2] ) ) && ( ip_addr[3] == '.' )
              && ( isdigit( ip_addr[4] ) ) && ( isdigit( ip_addr[5] ) ) && ( isdigit( ip_addr[6] ) ) && ( ip_addr[7] == '.' )
              && ( isdigit( ip_addr[8] ) ) && ( isdigit( ip_addr[9] ) ) && ( isdigit( ip_addr[10] ) ) && ( ip_addr[11] == '.' )
              && ( isdigit( ip_addr[12] ) ) && ( isdigit( ip_addr[13] ) ) && ( isdigit( ip_addr[14] ) ) ) )
        {
            printf( "Error, invalid IP address!\n" );
            return;
        }
        cfont631_enable_key_legend ( CFONT631_KEY_LEGEND_MINUS, CFONT631_KEY_LEGEND_PLUS,
                                     CFONT631_KEY_LEGEND_LEFT, CFONT631_KEY_LEGEND_RIGHT );
        cfont631_set_cursor_style( CFONT631_CURSOR_STYLE_BLINK_UNDER );
        cfont631_set_cursor_position( 1, 2 );
        cfont631_register_event_handler( handler, NULL );
        while ( 1 )
        {
            cfont631_write_xy( 1, 2, ip_addr, strlen(ip_addr) );
            fin = 0;
            while ( fin == 0 )
            {
                usleep( 10000 );
            }
            if ( fin == CFONT631_KEY_UP_LEFT )
            {
                if ( current_digit == 0 )
                {
                    printf( "cancel\n" );
                    break;
                }
                else
                {
                    // Subtract from current digit
                    current_octet = atoi( (char *)&ip_addr[current_digit - ( current_digit % 4 )] );
                    if ( ( current_digit % 4 ) == 1 ) current_octet -= 100;
                    else if ( ( current_digit % 4 ) == 2 ) current_octet -= 10;
                    else current_octet--;
                    if ( current_octet < 0 ) current_octet = 0;
                    sprintf( ascii_octet, "%03d", current_octet );
                    strncpy( (char *)&ip_addr[current_digit - ( current_digit % 4 )], ascii_octet, 3 );
                }
            }
            else if ( fin == CFONT631_KEY_UP_RIGHT )
            {
                if ( current_digit == 0 )
                {
                    printf( "%s\n", ip_addr );
                    break;
                }
                else
                {
                    // Add to current digit
                    current_octet = atoi( (char *)&ip_addr[current_digit - ( current_digit % 4 )] );
                    if ( ( current_digit % 4 ) == 1 ) current_octet += 100;
                    else if ( ( current_digit % 4 ) == 2 ) current_octet += 10;
                    else current_octet++;
                    if ( current_octet > 255 ) current_octet = 255;
                    sprintf( ascii_octet, "%03d", current_octet );
                    strncpy( (char *)&ip_addr[current_digit - ( current_digit % 4 )], ascii_octet, 3 );
                }
            }
            else if ( fin == CFONT631_KEY_DOWN_LEFT )
            {
                if ( current_digit == 0 ) current_digit = 16;
                current_digit--;
                if ( current_digit == 4 || current_digit == 8 || current_digit == 12 ) current_digit--;
                if ( current_digit == 0 )
                {
                    // Cursor is off of the IP; allow control functions
                    cfont631_enable_key_legend( CFONT631_KEY_LEGEND_CANCEL, CFONT631_KEY_LEGEND_CHECK,
                                                CFONT631_KEY_LEGEND_LEFT, CFONT631_KEY_LEGEND_RIGHT );
                    cfont631_set_cursor_style( CFONT631_CURSOR_STYLE_NO_CURSOR );
                }
                else
                {
                    // Cursor is on of the IP; allow decrement/increment functions
                    cfont631_enable_key_legend( CFONT631_KEY_LEGEND_MINUS, CFONT631_KEY_LEGEND_PLUS,
                                                CFONT631_KEY_LEGEND_LEFT, CFONT631_KEY_LEGEND_RIGHT );
                    cfont631_set_cursor_style( CFONT631_CURSOR_STYLE_BLINK_UNDER );
                    cfont631_set_cursor_position( current_digit, 2 );
                }
            }
            else if ( fin == CFONT631_KEY_DOWN_RIGHT )
            {
                current_digit++;
                if ( current_digit == 16 ) current_digit = 0;
                if ( current_digit == 4 || current_digit == 8 || current_digit == 12 ) current_digit++;
                if ( current_digit == 0 )
                {
                    // Cursor is off of the IP; allow control functions
                    cfont631_enable_key_legend( CFONT631_KEY_LEGEND_CANCEL, CFONT631_KEY_LEGEND_CHECK,
                                                CFONT631_KEY_LEGEND_LEFT, CFONT631_KEY_LEGEND_RIGHT );
                    cfont631_set_cursor_style( CFONT631_CURSOR_STYLE_NO_CURSOR );
                }
                else
                {
                    // Cursor is on of the IP; allow decrement/increment functions
                    cfont631_enable_key_legend( CFONT631_KEY_LEGEND_MINUS, CFONT631_KEY_LEGEND_PLUS,
                                                CFONT631_KEY_LEGEND_LEFT, CFONT631_KEY_LEGEND_RIGHT );
                    cfont631_set_cursor_style( CFONT631_CURSOR_STYLE_BLINK_UNDER );
                    cfont631_set_cursor_position( current_digit, 2 );
                }
            }
        }
        cfont631_disable_key_legend();
        cfont631_set_cursor_style( CFONT631_CURSOR_STYLE_NO_CURSOR );
        return;
    }

    printf( "\n" );
    printf( "cfa631 Error: Unknown command entered!\n\n" );

    return;
}

/*************************************************************************/
/**  MAIN                                                               **/
/*************************************************************************/

int main( int argc, char *argv[] )
{
    if ( argc <= 1 )
    {
        usage( argv );
        exit( 1 );
    }
    else
    {
        int fd;

        fd = cfont631_open( SERIAL_PORT );
        runcommands( argc, argv );
        cfont631_close();
        return( 0 );
    }
}

static void define_bignums (void) {
	const uint8 bignum[8][8] = {

	//   . . . . . *   001   * * * * * *   077
	//   . . . . . *   001   . . . . . *   001
	//   . . . . . *   001   . . . . . *   001
	//   . . . . . *   001   . . . . . *   001
	//   . . . . . *   001   . . . . . *   001
	//   . . . . . *   001   . . . . . *   001
	//   . . . . . *   001   . . . . . *   001
	//   . . . . . *   001   . . . . . *   001

	{ 001, 001, 001, 001, 001, 001, 001, 001 },
	{ 077, 001, 001, 001, 001, 001, 001, 001 },

	//   * * * * * *   077   * * * * * *   077
	//   * . . . . .   040   . . . . . *   001
	//   * . . . . .   040   . . . . . *   001
	//   * . . . . .   040   . . . . . *   001
	//   * . . . . .   040   . . . . . *   001
	//   * . . . . .   040   . . . . . *   001
	//   * . . . . .   040   . . . . . *   001
	//   * * * * * *   077   * * * * * *   077

	{ 077, 040, 040, 040, 040, 040, 040, 077 },
	{ 077, 001, 001, 001, 001, 001, 001, 077 },

	//   * . . . . *   041   * * * * * *   077
	//   * . . . . *   041   * . . . . .   040
	//   * . . . . *   041   * . . . . .   040
	//   * . . . . *   041   * . . . . .   040
	//   * . . . . *   041   * . . . . .   040
	//   * . . . . *   041   * . . . . .   040
	//   * . . . . *   041   * . . . . .   040
	//   * * * * * *   077   * . . . . .   040

	{ 041, 041, 041, 041, 041, 041, 041, 077 },
	{ 077, 040, 040, 040, 040, 040, 040, 040 },

	//   * * * * * *   077   * * * * * *   077
	//   * . . . . *   041   * . . . . *   041
	//   * . . . . *   041   * . . . . *   041
	//   * . . . . *   041   * . . . . *   041
	//   * . . . . *   041   * . . . . *   041
	//   * . . . . *   041   * . . . . *   041
	//   * . . . . *   041   * . . . . *   041
	//   * * * * * *   077   * . . . . *   041

	{ 077, 041, 041, 041, 041, 041, 041, 077 },
	{ 077, 041, 041, 041, 041, 041, 041, 041 }};

	for (int i = 0; i < 8; i++) {
	        cfont631_define_char( i, bignum[i] );
	}
}

static const int bignum_map_hi[10] = { 7, 0, 1, 1, 4, 5, 5, 1, 6, 6 };
static const int bignum_map_lo[10] = { 4, 0, 2, 3, 0, 3, 6, 0, 4, 0 };

static void bignum_write_digit_x( const int x, int num ) {

	if (0 > num || num > 9) return;

	cfont631_write_char_xy( x, 1, bignum_map_hi[num] );
	cfont631_write_char_xy( x, 2, bignum_map_lo[num] );
}

static void bignum_write_text_x( const int x, const char *s ) {
	char hi[128], lo[128];
	char *hip = hi, *lop = lo;
	const char *p = s;
	int len = 0;
	while (*p) {
		if (isdigit(*p)) {
			int d = *p++ - '0';
			*hip++ = bignum_map_hi[d];
			*lop++ = bignum_map_lo[d];
		} else {
			*hip++ = *p;
			*lop++ = *p++;
		}
		len++;
	}

	cfont631_write_xy( x, 1, hi, len );
	cfont631_write_xy( x, 2, lo, len );
}

/* eof */
